/// @addtogroup bitSet
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file CubCellSetT.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) Marian Mrozek 2005-2006
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.





#if !defined(_TOP_REP_SET_H_)
#define _TOP_REP_SET_H_
#include "bitSet/EuclBitSetT.h"
#include "repSet/RepSet.h"
#include <stack>

template <typename P_EuclSet>
class CubCellSetT : public P_EuclSet{
public:
  typedef typename P_EuclSet::Pixel Pixel;
  typedef typename P_EuclSet::BitCoordIterator BitCoordIterator;
  typedef typename P_EuclSet::BitIterator BitIterator;
  typedef typename P_EuclSet::PointIterator PointIterator;
  typedef typename P_EuclSet::PointCoordIterator PointCoordIterator;
  typedef typename P_EuclSet::NeighbIterator NeighbIterator;
  typedef typename P_EuclSet::Word Word;

  CubCellSetT():P_EuclSet(){};
  CubCellSetT(const int* A_w, bool clear=false):P_EuclSet(A_w,clear){};  // BitmapT with w[i] pixels in dimension i
  CubCellSetT(const CubCellSetT& A_org, bool clear=false);
  CubCellSetT(int size, bool clear=false):P_EuclSet(size,clear){};       // BitmapT with size pixels in each dimension
  CubCellSetT(const P_EuclSet& A_EuclSet);
  CubCellSetT(const RepSet<ElementaryCube>& A_RepSetOfElementaryCube);

  // Conversion to the full cubical set obtained
  // by treating each cell as a full cube
  operator P_EuclSet&(){
    return *reinterpret_cast<P_EuclSet* >(this);
  }


  void rescale(int A_factor);
  void rescale(const int A_factor[]);

  // Verify if the given pixel represents a vertex in the
  // representalbe set, i.e. if it is degenerate in all dimensions

  // Iterator version
  bool isVertex(BitCoordIterator& A_it) const;

  // Verify if the given pixel represents a maximal cell in the
  // representalbe set, i.e. if it is not a face of a cell of dimesion one greater
  bool isMaximal(BitCoordIterator& A_it) const;

  // Verify if the given pixel represents a free cell in the representable set,
  // i.e. a cell which is a face of exaclty one cell of dimension one greater
  // (codimension one coface)
  // If true, return the codimension one coface
  bool isFreeFace(BitCoordIterator& A_faceIt, BitCoordIterator& A_coFaceIt) const;

  // Verify if the given pixel represents a cofree cell in the representable set,
  // i.e. a cell which has exaclty one face of dimension one lower in the representable set
  // (codimension one face)
  // If true, return the codimension one face

  bool findCoFreeFaceOrFreeVertex(const CubCellSetT<P_EuclSet>& source,PointIterator& A_startIter,BitCoordIterator& A_it)const;

  // Iterator version
  bool isFreeCoFace(BitCoordIterator& A_it, BitCoordIterator& A_codimOneFaceIt) const;


  // Iterator version
  void getCoFaces(BitCoordIterator& A_it,std::vector<BitCoordIterator>& A_codimOneCofacesIterators) const;

  // Find all codim one faces of a given pixel
  void getFaces(BitCoordIterator& A_it,std::vector<BitCoordIterator>& A_facesIterators) const;

  // A specialized version of shave for representable sets
  void shave();
  // A systematic version of removeTrivialPairs() below
  int reduce();
  // A systematic version of removeCoTrivialPairs() below
  int coReduce();       // iterator version

  // A speeded up version for compact sets
  bool findFreeCoFace(PointIterator& A_startIter,BitCoordIterator& A_it) const;
  bool findVertex(PointIterator& A_startIter,BitCoordIterator& A_it) const;
  int coReduceCompactSet();
  int coReduceCompactSetFast();

  // Auxiliary, recursive method for cubicalClosureSkeleton
  bool belongsToCubicalClosure(BitCoordIterator& A_it);

  // Given a CubRepSet consisting of vertices only, compute
  // the A_skeletonDimension skeleton of the cubical closure
  // defined by the recursive formula that a cube belongs to the
  // closure if all it faces belong to the closure
  void cubicalClosureSkeleton(int A_skeletonDimension);

  friend void swap(CubCellSetT& A_set1,CubCellSetT& A_set2){
    swap(*static_cast<P_EuclSet*>(&A_set1),*static_cast<P_EuclSet*>(&A_set2));
  }
};

// -------------------------------------------------------------------------------------- //



#endif // _TOP_REP_SET_H_
/// @}

