/// @addtogroup multiEngHom
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file AsHomMD.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) Marian Mrozek 2005-2006
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_AS_HOM_T_H_)
#define _AS_HOM_T_H_

#include <iostream>
#include <fstream>
#include <sstream>
#include <exception>
#include <new>
using namespace std;

#include "vectalg/matrixSlice.h"
#include "matrixAlgorithms/intMatrixAlgorithms.hpp"

#include "homologicalAlgebra/embeddingDim.h"
#include "auxil/ofstreamcout.h"
#include "auxil/Stopwatch.h"
#include "auxil/memSize.h"
#include "auxil/CRef.h"
#include "auxil/stringOstream.h"

ofstreamcout fcout;

using namespace capd;
using namespace matrixAlgorithms;
using namespace std;


#include "homologicalAlgebra/homologicalAlgebra.hpp"
#include "homologicalAlgebra/readCubFile.hpp"
#include "homologicalAlgebra/homAlgFunctors.hpp"
#include "homologicalAlgebra/cubSetFunctors.hpp"
#include "homologicalAlgebra/acyclicConfigs.hpp"
#include "bitSet/EmbDimException.h"

template <int dim>
struct AsHom{
  typedef CubSetT<EuclBitSetT<BitSetT<BitmapT<unsigned long int> >,dim> > BCubSet;
  typedef CubCellSetT<EuclBitSetT<BitSetT<BitmapT<unsigned long int> >,dim> > BCubCelSet;
  typedef CRef<BCubSet> BCubSetCR;
  typedef CubSetFunctors<BCubSet,BCubCelSet> CubSetFunct;
  typedef CRef<HomologySignature> (*HomologyAlgorithm)(CRef<BCubSet>);

  static AsHom selector;

  std::map<std::string,HomologyAlgorithm> homologyAlgorithms;

  HomologyAlgorithm setupHomologyAlgorithm(const std::string& A_engine){
    HomologyAlgorithm selectedAlgorithm;
    if(homologyAlgorithms.count(A_engine)) selectedAlgorithm=homologyAlgorithms[A_engine];
    else{
      std::string s;
      s << "Bad engine: " << A_engine << " not known or not supported for this dimension\n";
      throw EmbDimException(s);
    }
    return selectedAlgorithm;
  }

  void showResults(const CRef<HomologySignature>& A_hsCR,const std::string& A_engine, int A_verbose){
    fcout.turnOn();
    if(A_verbose) fcout << "AsHomMD 0.9 (c) Marian Mrozek, 2006" << std::endl;
    fcout << std::endl;
    switch(A_verbose){
      case 0: for(unsigned int i=0;i<=A_hsCR().size();++i) fcout << A_hsCR().bettiNumber(i) << " "; break;
      case 1: for(unsigned int i=0;i<=A_hsCR().size();++i){
                fcout << "Betti[" << i << "]=" << A_hsCR().bettiNumber(i) << std::endl;
              }
              break;
      default: fcout << A_hsCR();
    }
    fcout << std::endl;
  }
  // This method runs the particular homology engine for the cubical set in the prescribed file
  // in the prescribed dimension given as the dim template parameter
  void operator()(const char* A_fileName, const std::string& A_engine, int A_verbose){
    if(A_verbose>=3) fcout.turnOn();
    else fcout.turnOff();

    CRef<BCubSet> cubSetCR;
    try{
      cubSetCR=readCubFile<BCubSet>(A_fileName);
    }catch(EmbDimException){
      throw EmbDimException("Incorrect embDim");
    }
    if(cubSetCR().embDim()!=dim) throw EmbDimException("Incorrect embDim");

    HomologyAlgorithm selectedAlgorithm=setupHomologyAlgorithm(A_engine);
    Stopwatch swTot;
    CRef<HomologySignature> hsCR=selectedAlgorithm(cubSetCR);
    hsCR().shrink();
    showResults(hsCR,A_engine,A_verbose);
    if(A_verbose) fcout << "Computation completed in " << swTot << " under " << A_engine << " engine." << std::endl;
  }

  void operator()(int embDim, const int dims[], const char* bytes, int* betti, const std::string& A_engine="CR", int A_verbose=0){
    if(A_verbose>=3) fcout.turnOn();
    else fcout.turnOff();

    HomologyAlgorithm selectedAlgorithm=setupHomologyAlgorithm(A_engine);
    Stopwatch swTot;
    CRef<BCubSet> cubSetCR(new BCubSet(dims,bytes));
    CRef<HomologySignature> hsCR=selectedAlgorithm(cubSetCR);
    hsCR().shrink();
    for(int i=0;i<embDim;++i) betti[i]=hsCR().bettiNumber(i);
    if(A_verbose) fcout << "Computation completed in " << swTot << " under " << A_engine << " engine." << std::endl;
  }

}; // AsHom<dim>

template<int dim>
AsHom<dim> AsHom<dim>::selector;



template <>
struct AsHom<0>{
  AsHom<2>& ch2;
  AsHom<3>& ch3;
  AsHom<4>& ch4;

  typedef AsHom<2>::BCubSet BCubSet2;
  typedef AsHom<3>::BCubSet BCubSet3;
  typedef AsHom<4>::BCubSet BCubSet4;

  typedef AsHom<2>::CubSetFunct CubSetFunct2;
  typedef AsHom<3>::CubSetFunct CubSetFunct3;
  typedef AsHom<4>::CubSetFunct CubSetFunct4;

  typedef AsHom<2>::HomologyAlgorithm HomologyAlgorithm2;
  typedef AsHom<3>::HomologyAlgorithm HomologyAlgorithm3;
  typedef AsHom<4>::HomologyAlgorithm HomologyAlgorithm4;

  AsHom():
    ch2(AsHom<2>::selector),
    ch3(AsHom<3>::selector),
    ch4(AsHom<4>::selector)
  {
    ch2.homologyAlgorithms.insert(std::make_pair("ASHns",&CubSetFunct2::HomologyViaAcyclicSubspaceHOM));
    ch2.homologyAlgorithms.insert(std::make_pair("ASH",&CubSetFunct2::HomologyViaAcyclicSubspaceHOMShaved));
    ch2.homologyAlgorithms.insert(std::make_pair("ASns",&CubSetFunct2::HomologyViaAcyclicSubspaceSI));
    ch2.homologyAlgorithms.insert(std::make_pair("AS",&CubSetFunct2::HomologyViaAcyclicSubspaceSIShaved));
    ch2.homologyAlgorithms.insert(std::make_pair("ASRns",&CubSetFunct2::HomologyViaAcyclicSubspaceSIR));
    ch2.homologyAlgorithms.insert(std::make_pair("ASR",&CubSetFunct2::HomologyViaAcyclicSubspaceSIRShaved));

    ch3.homologyAlgorithms.insert(std::make_pair("ASHns",&CubSetFunct3::HomologyViaAcyclicSubspaceHOM));
    ch3.homologyAlgorithms.insert(std::make_pair("ASH",&CubSetFunct3::HomologyViaAcyclicSubspaceHOMShaved));
    ch3.homologyAlgorithms.insert(std::make_pair("ASns",&CubSetFunct3::HomologyViaAcyclicSubspaceSI));
    ch3.homologyAlgorithms.insert(std::make_pair("AS",&CubSetFunct3::HomologyViaAcyclicSubspaceSIShaved));
    ch3.homologyAlgorithms.insert(std::make_pair("ASRns",&CubSetFunct3::HomologyViaAcyclicSubspaceSIR));
    ch3.homologyAlgorithms.insert(std::make_pair("ASR",&CubSetFunct3::HomologyViaAcyclicSubspaceSIRShaved));
    ch3.homologyAlgorithms.insert(std::make_pair("ASLTns",&CubSetFunct3::HomologyViaAcyclicSubspaceLTD3));
    ch3.homologyAlgorithms.insert(std::make_pair("ASLT",&CubSetFunct3::HomologyViaAcyclicSubspaceLTD3Shaved));

    ch4.homologyAlgorithms.insert(std::make_pair("ASHns",&CubSetFunct4::HomologyViaAcyclicSubspaceHOM));
    ch4.homologyAlgorithms.insert(std::make_pair("ASH",&CubSetFunct4::HomologyViaAcyclicSubspaceHOMShaved));
    ch4.homologyAlgorithms.insert(std::make_pair("ASns",&CubSetFunct4::HomologyViaAcyclicSubspaceSI));
    ch4.homologyAlgorithms.insert(std::make_pair("AS",&CubSetFunct4::HomologyViaAcyclicSubspaceSIShaved));
    ch4.homologyAlgorithms.insert(std::make_pair("ASRns",&CubSetFunct4::HomologyViaAcyclicSubspaceSIR));
    ch4.homologyAlgorithms.insert(std::make_pair("ASR",&CubSetFunct4::HomologyViaAcyclicSubspaceSIRShaved));
  }

  void failure(const std::string& A_engine, int A_verbose){
    if(A_engine=="ASLT" || A_engine=="ASLTns"){
      string s;
      s << "The only dimension accepted for " << A_engine << " is 3\n";
      throw EmbDimException(s);
    }else{
      string s;
      s << A_engine << " is incorrect or not available for this dimension.\n";
      throw EmbDimException(s);
    }
  }

  // This method runs the particular homology engine for the cubical set in the prescribed file
  // trying various dimensions
  void operator()(const char* A_fileName, const std::string& A_engine, int A_verbose){
    readAcyclicConfigs();
    for(int i=2;i<5;++i){
      try{
        switch(i){
          case 2: ch2(A_fileName,A_engine,A_verbose);
                  return;
          case 3: ch3(A_fileName,A_engine,A_verbose);
                  return;
          case 4: ch4(A_fileName,A_engine,A_verbose);
                  return;
        }
      }catch(EmbDimException& e){
        if(e.what() != std::string("Incorrect embDim")) throw;
      };
    }
    failure(A_engine, A_verbose);
  }

  // This method runs the particular homology engine for the cubical set in the provided bitmap
  // trying various dimensions
  void operator()(int embDim, const int dims[], const char* bytes, int* betti, const std::string& A_engine="CR", int A_verbose=1){
    readAcyclicConfigs();
    if(dims[0] % 32) throw std::runtime_error("The 0th dimension must be a multiciplity of 32!");
    CRef<BCubSet> cubSetCR(new BCubSet(dims,bytes));
    switch(embDim){
      case 2: ch2(2,dims,bytes,betti,A_engine,A_verbose);
              return;
      case 3: ch3(3,dims,bytes,betti,A_engine,A_verbose);
              return;
      case 4: ch4(4,dims,bytes,betti,A_engine,A_verbose);
              return;
    }
  }

};// AsHom<0>

#endif //_AS_HOM_T_H_
/// @}

