/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file FreeChainComplex.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_FREECHAINCOMPLEX_H_)
#define _FREECHAINCOMPLEX_H_

template<typename freeModuleType>
class FreeChainComplex;

template<typename freeModuleType>
std::ostream& operator<<(std::ostream& out,const FreeChainComplex<freeModuleType>&);

template<typename freeModuleType>
void swap(FreeChainComplex<freeModuleType>& A_fcc1,FreeChainComplex<freeModuleType>&);

template<typename freeModuleType>
class ReducibleFreeChainComplex;

// ********** class FreeChainComplex ********** //
/*
  This class is used to store a free, finitely generated chain complex.
  This is not a dynamic structure as ReducibleFreeChainComplex.
  Its objects are not allowed to change, once constructed.

  The class stores a vector of free modules and a vector of boundary homomorphisms
  for the interval of grades, where the modules are non trivial.

  The two constructors implemented in FreeChainComplex.hpp
  are described in detail there
*/
template<typename freeModuleType>
class FreeChainComplex{
public:
  typedef typename freeModuleType::MatrixType MatrixType;
  typedef typename freeModuleType::GeneratorType GeneratorType;

  FreeChainComplex(){};
  // construct a free chain complex from a std::vector of generators - see FreeChainComplex.hpp
  // needed only in some conretizations, so changed to template to avoid forced compilation
  template<typename GeneratorType2>
  FreeChainComplex(const std::vector<GeneratorType2>& A_gens);
  // construct a free chain complex from a std::set of generators - see FreeChainComplex.hpp
  // needed only in some conretizations, so changed to template to avoid forced compilation
  template<typename GeneratorType2>
  FreeChainComplex(const std::set<GeneratorType2>& A_gensSet);

  const freeModuleType& chainModule(int i) const{
    return m_chainModule[i];
  }
  const FreeModuleHomomorphism<GeneratorType,MatrixType>& boundaryHomomorphism(int i) const{
    return m_boundaryHomomorphism[i];
  }
  int topDim() const{
    return m_topDim;
  }
  friend void swap<freeModuleType>(FreeChainComplex<freeModuleType>& A_fcc1,FreeChainComplex<freeModuleType>& A_fcc2);
  friend std::ostream& operator<< <freeModuleType>(std::ostream& out, const FreeChainComplex& A_fCC);


  template<typename freeModuleType2>
  friend class ReducibleFreeChainComplex;

private:
  int m_topDim;
  std::vector<FreeModule<GeneratorType,MatrixType> > m_chainModule;
  std::vector< FreeModuleHomomorphism<GeneratorType,MatrixType> > m_boundaryHomomorphism;
}; // ********** class FreeChainComplex ********** //

// ++++++++++++++++++++++++++++ inline functions definitions +++++++++++++++++++++++++++++++++++ //

template<typename freeModuleType>
inline void swap(FreeChainComplex<freeModuleType>& A_fcc1,FreeChainComplex<freeModuleType>& A_fcc2){
  std::swap(A_fcc1.m_topDim,A_fcc2.m_topDim);
  std::swap(A_fcc1.m_chainModule,A_fcc2.m_chainModule);
  std::swap(A_fcc1.m_boundaryHomomorphism,A_fcc2.m_boundaryHomomorphism);
}


#endif //_FREECHAINCOMPLEX_H_
/// @}

