/// @addtogroup homologicalAlgebra
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file ReducibleFreeChainComplex.h
///
/// @author Marian Mrozek
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2006 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_REDUCIBLEFREECHAINCOMPLEX_H_)
#define _REDUCIBLEFREECHAINCOMPLEX_H_

// ********** class ReducibleFreeChainComplex ********** //

/*
 The class ReducibleFreeChainComplex in constrast to FreeChainComplex
 is designed for homology computations via reductions described in
    T. Kaczynski, M. Mrozek and M. \'Slusarek,
    Homology Computation by Reduction of Chain Complexes,
    Computers & Mathematics , 35(1998), no. 4, 59--70
  and
    M. Mrozek, K. Mischaikow and T. Kaczynski,
    Computational Homology,
    Applied Mathematical Sciences 157, Springer-Verlag, New York, 2004
  Therefore, unlike the objects of FreeChainComplex, the objects of
  ReducibleFreeChainComplex may change during computations.

   The chain complex is stored in the data structure

     std::vector<std::map<GeneratorCode,std::map<GeneratorCode,ScalarType> > >

   where GeneratorCode is typedef to int,
   vector goes through all grades of the chain complex,
   the external map goes through all generators for a given grade
   and the internal map stores the boundary of this generator as pairs (generator,coefficient)
   for non-zero coefficients

  In future additions pointers to generators may be added, but for the moment
  only generator codes are used, which means that outcome of computations
  cannot present generators of homology.


*/

#include <list>
#include <queue>
#include "homologicalAlgebra/ChainT.h"

template<typename freeModuleType2>
void swap(ReducibleFreeChainComplex<freeModuleType2>& A_rfcc1,ReducibleFreeChainComplex<freeModuleType2>& A_rfcc2);

template<typename freeModuleType2>
std::ostream& operator<<(std::ostream& out,const ReducibleFreeChainComplex<freeModuleType2>& A_rfc);

template<typename freeModuleType>
class ReducibleFreeChainComplex{
  public:
    typedef typename freeModuleType::GeneratorType GeneratorType;
    typedef typename freeModuleType::MatrixType MatrixType;
    typedef typename freeModuleType::MatrixType::ScalarType ScalarType;
    typedef int GeneratorCode;
    typedef ChainT<std::map<GeneratorCode,ScalarType> > Chain;
    typedef Chain CoChain;
    typedef typename std::map<GeneratorCode,Chain> GeneratorBoundaries;
    typedef typename std::map<GeneratorCode,CoChain> GeneratorCoBoundaries;
    typedef typename std::vector<GeneratorBoundaries> GradedGeneratorBoundaries;
    typedef typename std::vector<GeneratorCoBoundaries> GradedGeneratorCoBoundaries;
    typedef int (ReducibleFreeChainComplex::*ReducibleFreeChainComplex_void_Ptr)();

    // -------------------------------------------------------------------------------------- //

    ReducibleFreeChainComplex(){};

    /*
        This constructor constructs a reducible free chain complex from a chain complex
    */
    template<typename freeModuleType2>
    ReducibleFreeChainComplex(const FreeChainComplex<freeModuleType2>& fcc);

    /*  -------------------------------------------------------------------------------------- //
        This constructor takes a vector of generators as the argument.
        The generators are assumed to provide the method boundary(),
        returning the map mapping primary faces to their coefficients
        in the boundary
        Faces of all generators are assumed to be generators too.
        They are added to the list of generators automatically, so there is no need to add them
        to the input vector
    */

    template<typename GeneratorType2>
    ReducibleFreeChainComplex(
      const std::vector<GeneratorType2>& A_gens // vector of generators
    );

    /* -------------------------------------------------------------------------------------- //
      This constructor takes a set of generators as the argument.
      The generators are assumed to know all their faces by providing the method primaryFaces()
      returning the vector of primary faces
      as well as the method boundary() returning the map mapping primary faces to their coefficients
      in the boundary
      Faces of generators are not automatically assumed to be generators in the chain complex, so
      all generators must be listed.
      The algebraic boundary in the chain module is taken to consists only of the faces present
      in the provided set. For this reason (search), set is taken here as the proper data structure
      on input
      No check is made if the provided set of generators leads to a correct chain complex!!! - should be changed!!!
    */

    template<typename GeneratorType2>
    ReducibleFreeChainComplex(
      const std::set<GeneratorType2>& A_gensSet // set of generators
    );

    // ------------------------------------------------------------ //
    operator FreeChainComplex<FreeModule<GeneratorCode,Matrix<int,0,0> > >();

    // ------------------------------------------------------------ //
    int topDim();


    // ------------------------------------------------------------ //
    void reducePair(
      int q, // grade in which the reduction is performed
      // iterator pointing to the first element of the reduction pair (b,a) or (coFace,face)
      // pointing to the external map
      typename GeneratorBoundaries::iterator coFacePos,
      // iterator pointing to the second element of the reduction pair (b,a) or (coFace,face)
      // pointing to the map storing the boundary of generator b
      typename Chain::iterator facePos
    );

    // ------------------------------------------------------------ //
    // This is a direct implementation of the KMS algorithm
    int reduce();

    // ------------------------------------------------------------ //
    // This measures the amount of work needed to make the change of the boundary operator when reducing
    int findMeasure(const int& coFaceDim,const GeneratorCode& coFace, const GeneratorCode& face);

    // ------------------------------------------------------------ //
    // A class auxiliary for reduceViaSort below
    friend class ReduciblePair;
    class ReduciblePair;

    void extractReduciblePairsInVicinityOf(
      int dim,
      GeneratorCode g,
      std::set<ReduciblePair>& source,
      std::list<ReduciblePair>& destination
    );

    // ------------------------------------------------------------ //
    // Here we first search for all the reducible pairs and sort them
    // according to the measure defined above. Then we reduce the reducible
    // pairs starting from the pair with the lowest measure
    // We keep updating the measure and position in the priority queue
    // after every reduction
    int reduceViaSort();

    // ------------------------------------------------------------ //
    // Find the first reducible pair, reduce it, search for new reducible
    // pairs in the vicinity of the reduced coFace, choose the one with smallest
    // measure and proceed this way until no reducible pairs are found
    int reduceViaLocalSort();

    // ------------------------------------------------------------ //
    // Search if a reducible pair of any type exists, if so return true
    // reduce all reducible pairs of the prescribed coboundary size
    // It may happen that no pairs are reduced, but some reducible pairs exist
    bool reduceViaStack(unsigned int searchedCoboundrySize, int& pairsReduced);

    // ------------------------------------------------------------ //
    void reduceFromBottom();

    // ------------------------------------------------------------ //
    // Generators, which have zero boundary and coboundary are simple homology generators
    // All these generators can be removed from the complex
    // What remains is still a complex
    // The total homology is the homology given by the simple homology generators plus the homology of
    // what is left.
    // Return true if all generators were removed
    bool removeSimpleHomologyGenerators(std::vector<std::vector<GeneratorCode> >& simpleHomologyGenerators);

    // ------------------------------------------------------------ //
    friend void swap<freeModuleType>(ReducibleFreeChainComplex<freeModuleType>& A_rfcc1,ReducibleFreeChainComplex<freeModuleType>& A_rfcc2);
    friend std::ostream& operator<< <freeModuleType>(std::ostream& out,const ReducibleFreeChainComplex<freeModuleType>& A_rfc);

  // ------------------------------------------------------------ //
  private:
    GradedGeneratorBoundaries m_ggb;
    GradedGeneratorCoBoundaries m_ggCb;
};//ReducibleFreeChainComplex

// ++++++++++++++++++++++++++++ inline functions definitions +++++++++++++++++++++++++++++++++++ //

template<typename freeModuleType>
int ReducibleFreeChainComplex<freeModuleType>::topDim(){
  return m_ggb.size()-1;
}

// ------------------------------------------------------------ //

template<typename freeModuleType2>
inline void swap(ReducibleFreeChainComplex<freeModuleType2>& A_rfcc1,ReducibleFreeChainComplex<freeModuleType2>& A_rfcc2){
  std::swap(A_rfcc1.m_ggb,A_rfcc2.m_ggb);
  std::swap(A_rfcc1.m_ggCb,A_rfcc2.m_ggCb);
}

// ++++++++++++++++++++++++++++ Internal class definition +++++++++++++++++++++++++++++++++++++ //

template<typename freeModuleType>
class ReducibleFreeChainComplex<freeModuleType>::ReduciblePair{
  int coFaceDim;
  GeneratorCode coFace,face;
  int measure;

  public:
  ReduciblePair(int A_coFaceDim,GeneratorCode A_coFace,GeneratorCode A_face,int A_measure):
    coFaceDim(A_coFaceDim),coFace(A_coFace),face(A_face),measure(A_measure){}

  friend bool operator<(const ReduciblePair& p1,const ReduciblePair& p2){
    if(p1.measure !=  p2.measure) return p1.measure < p2.measure;
    if(p1.coFace  !=  p2.coFace) return p1.coFace  <  p2.coFace;
    return p1.face  <  p2.face;
  }
  friend bool operator>(const ReduciblePair& p1,const ReduciblePair& p2){
    if(p1.measure !=  p2.measure) return p1.measure > p2.measure;
    if(p1.coFace  !=  p2.coFace) return p1.coFace  >  p2.coFace;
    return p1.face  >  p2.face;
  }
  const GeneratorCode& CoFace(){
    return coFace;
  }
  const GeneratorCode& Face(){
    return face;
  }
  const int& CoFaceDim(){
    return coFaceDim;
  }
  void setMeasure(int m){
    measure=m;
  }
};

// ------------------------------------------------------------ //

#endif //_REDUCIBLEFREECHAINCOMPLEX_H_
/// @}

