/// @addtogroup interval
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file interval.h
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

//--------------
//               header file of class 'interval'
//--------------
#include <iostream>
#include <sstream>
#include <stdexcept>

#include "capd/power.h"
#include "capd/minmax.h"
#include "interval/intervalSetting.h"
#include "interval/rounding.h"

#ifndef _INTERVAL_CLASS
#define _INTERVAL_CLASS

#define EXP_TAYLOR 20
// Taylor series length used in the EXP function
#define SIN_TAYLOR 11
// Taylor series length used in the SIN function, it MUST be an ODD value
#define LOG_TAYLOR 40
// Taylor series length used in the LOG function, it MUST be an EVEN value

typedef double bound_type;  // the  type used in the definition interval class
                            // each end of an interval is of this type

class  interval;
extern const interval INV_E;
extern const interval pi;
extern const interval pi3by2;
extern const interval piby2;
extern const interval pi2;

class interval{

public:

   interval() {};
   interval(bound_type _value) : left(_value), right(_value) {};
   interval(bound_type _left,bound_type _right);
   interval(const interval& iv);

   interval& operator = (const interval& iv);
   interval& operator +=(const interval& iv);
   interval& operator -=(const interval& iv);
   interval& operator *=(const interval& iv);
   interval& operator /=(const interval& iv);

   bool contains(const double x) const;
   bool subset(const interval &iv) const;
   bool subsetInterior(const interval &iv) const;
   void split(interval &r);
   interval mid(void) const;

   friend int operator ==(const interval& iv1, const interval& iv2);
   friend int operator <=(const interval& iv1, const interval& iv2);
   friend int operator >=(const interval& iv1, const interval& iv2);
   friend int operator < (const interval& iv1, const interval& iv2);
   friend int operator > (const interval& iv1, const interval& iv2);
   friend int operator !=(const interval& iv1, const interval& iv2);
   friend int in(bound_type _value,const interval& iv);  // inclusion
   friend int in(const interval& small_iv ,const interval& iv,int interior);// inclusion


   friend interval operator -(const interval& iv);

   friend interval operator +(const interval& iv1, const interval& iv2);
   friend interval operator +(const interval& iv, const bound_type _val  );
   friend interval operator +(const bound_type _val, const interval& iv );
   friend interval operator -(const interval& iv1, const interval& iv2);
   friend interval operator -(const interval& iv, const bound_type _val  );
   friend interval operator -(const bound_type _val, const interval& iv );
   friend interval operator *(const interval& iv1, const interval& iv2);
   friend interval operator *(const interval& iv, const bound_type _val  );
   friend interval operator *(const bound_type _val, const interval& iv );
   friend interval operator /(const interval& iv1, const interval& iv2);
   friend interval operator %(const interval& iv1, const interval& iv2);
   friend interval operator %(const interval& iv1, const bound_type iv2);
   friend interval operator ^(const interval& iv1, const int i);
//    friend interval operator &(const interval&, const interval&);

   const bound_type& left_b(void) const { return left; }
   const bound_type& right_b(void) const { return right; }

   friend interval Left(const interval &iv);
   friend interval Right(const interval &iv);

   friend interval diameter(const interval &iv);
   friend interval nonnegativePart(const interval &iv);

   friend interval mid(const interval &iv);    // the middle point of interval
   friend interval ball(const interval &iv,const interval &r);
   friend void iSplit(interval *iv,double *diam);

   friend interval sin(const interval&);
   friend interval cos(const interval&);
   friend interval tan(const interval&);
   friend interval cot(const interval&);
   friend interval sinh(const interval&);
   friend interval cosh(const interval&);
   friend interval tanh(const interval&);
   friend interval coth(const interval&);
   friend interval exp(const interval&);
   friend interval log(const interval&);

   friend double fabs(const interval&);
   friend interval power(const interval&, int);
   friend interval power(const interval&, const interval&);
   friend interval sqrt(const interval&);
   friend interval Ihull(const interval & , const interval & );
   friend interval solve_affine_inclusion(const interval &a, const interval & p, const interval& c);
   friend interval solve_affine_inclusion(const interval &a, const interval & p, const interval& c, int &dir);

   friend std::ostream& operator<<(std::ostream &stream, const interval& iv);
   friend std::istream& operator>>(std::istream &stream, interval& iv);

#ifdef KRAK
   friend frstring &operator<<(frstring &f, const interval &i);
   friend frstring &operator>>(frstring &f, interval &i);

   friend Frame &operator<<(Frame &frame, const interval &iv);
#endif
private:

   bound_type  left;
   bound_type  right;
};

//############################ templates for input-output #######################


void interval_error(char *c, double left, double right);
void interval_test(interval &iv, char *comment);

//------------'inline' function for the class 'interval'

inline interval::interval(bound_type _left,bound_type _right):
       left(_left),right(_right)
{
#if defined(KRAK) || defined(DEBUGGING)
   if (left > right)
      interval_error("Interval constructor error!",left,right);
#endif
}

// a copying constructor
inline interval::interval(const interval& iv)
{
   left = iv.left;
   right = iv.right;
}

inline bool interval::contains(const double x) const
{
   return (left <= x) && (x<=right);
}

inline int Degenerate(interval &iv)
{
   return iv.contains(0.0);
}

inline int Degenerate(double &r)
{
   return (r==0.0);
}

inline bool interval::subset(const interval &iv) const
{
   return (left >= iv.left) && (right<=iv.right);
}

inline bool interval::subsetInterior(const interval &iv) const
{
   return (left > iv.left) && (right<iv.right);
}

inline interval interval::mid(void) const
{
   return((interval(right)+interval(left))/2);
}

//the relation  ==
inline int operator ==(const interval& iv1, const interval& iv2)
{
   return( (iv1.left==iv2.left) && (iv1.right==iv2.right));
}

// weak inequality relation <=
inline int operator <=(const interval& iv1, const interval& iv2)
{
   return(iv1.right<=iv2.left);
}

// strong inequality <
inline int operator <(const interval& iv1, const interval& iv2)
{
   return(iv1.right<iv2.left);
}

//weak inequality relation >=
inline int operator >=(const interval& iv1, const interval& iv2)
{
   return(iv1.left>=iv2.right );
}

//strong inequality relation >
inline int operator >(const interval& iv1, const interval& iv2)
{
   return(iv1.left>iv2.right);
}

//the relation  !=  ( as a negation of  ==)
inline int operator !=(const interval& iv1, const interval& iv2)
{
   return(iv1.left!=iv2.left || iv1.right!=iv2.right);
}

//a different operator for powers
inline interval operator^(const interval& value, int ile)
{
   return power(value, ile);
}

// sending an interval to a stream
inline std::ostream& operator <<(std::ostream& s, const interval& iv)
{
   #ifndef UP_ROUNDING
   round_nearest();
   #endif
   return(s << "[" << iv.left << "," << iv.right << "]");
}

#ifdef KRAK
inline frstring &operator<<(frstring &f, const interval& iv)
{
   #ifndef UP_ROUNDING
   round_nearest();
   #endif
   f << "[" << iv.left << "," << iv.right << "]";
   return f;
}
#endif

namespace capd{
//maximum
inline interval max(const interval& iv1, const interval& iv2)
{
   return interval((iv1.left_b()>iv2.left_b() ? iv1.left_b() : iv2.left_b()),
                  (iv1.right_b()>iv2.right_b() ? iv1.right_b() : iv2.right_b()));
}

//minimum
inline interval min(const interval& iv1, const interval& iv2)
{
   return interval((iv1.left_b()<iv2.left_b() ? iv1.left_b() : iv2.left_b()),
                  (iv1.right_b()<iv2.right_b() ? iv1.right_b() : iv2.right_b()));
}
}
//Unary minus
inline interval operator-(const interval& iv)
{
   return(interval(-iv.right,-iv.left));
}



// Diameter
inline interval diameter(const interval& iv)
{
   return(interval(iv.right)-interval(iv.left));
}

//midpoint
inline interval mid(const interval& iv)
{
   return((interval(iv.right)+interval(iv.left))/2);
}

//nonnegative part
inline interval nonnegativePart(const interval& iv)
{
   if(iv.right_b()>=0)
      return interval(capd::max(0.,iv.left_b()), iv.right_b() );
   else
   {
      std::ostringstream message;
      message << "nonnegative part of " << iv << " is empty\n";
      throw std::runtime_error(message.str());
      return interval(0);
   }
}


inline interval ball(const interval& iv,const interval &r)
{
   return interval( (iv-r).left_b(),(iv+r).right_b() );
}

//Number inclusion _value \in iv
inline int in(bound_type _value,const interval& iv){
   return ( _value >= iv.left && _value <= iv.right );
}

//Interval inclusion
//  returns 1 if small_iv \subset \interior in, if interior==1
//          1 if small_iv \subset \interior in, if interior==0
//   0 - otherwise
inline int in(const interval& small_iv ,const interval& iv,int interior)
{
   if (!interior)
     return ( small_iv.left >= iv.left && small_iv.right <= iv.right );
   else
     return ( small_iv.left > iv.left && small_iv.right < iv.right );
}//--------------------------------------------------------------------


int  intersection(interval x1, interval x2,interval &cp);
// on output:  if common part of <x1> and <x2> is nonempty then
//                cp =intersection of <x1> and <x2>
//                 and 1 is returned
//              otherwise 0 is returned and cp is untouched

inline interval Left(const interval &iv){
  return interval(iv.left);
}

inline interval Right(const interval &iv){
  return interval(iv.right);
}

inline double Left(double d){
  return d;
}

inline double Right(double d){
  return d;
}



#if defined(KRAK) || defined(DEBUGGING)
#  include "krak/krak.h"



inline Frame &operator<<(Frame &frame,const interval &iv){
  #ifndef UP_ROUNDING
    round_nearest();
  #endif
  frame << "[" << iv.left_b() << "," << iv.right_b() << "]";
  return frame;
}


#endif

#endif // _INTERVAL_CLASS

/// @}
