/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file dynamicContainer.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2005 */

#ifndef _DynamicContainer_H
#define _DynamicContainer_H

#include <cstdlib> // for size_t
#include <new>

namespace capd{
namespace vectalg{

template<typename Scalar>
class Container<Scalar,0>
{
public:
   typedef Scalar ScalarType;
   typedef ScalarType* iterator;
   typedef const ScalarType* const_iterator;

   Container();
   explicit Container(int);
   Container(int,bool); // it does not insert zeros
   Container(const Container&);
   ~Container();

   iterator begin();
   iterator end();
   const_iterator begin() const;
   const_iterator end() const;

   Container& operator=(const Container&);
   void resize(int);

   ScalarType& operator[](int);
   const ScalarType& operator[](int) const;
   ScalarType& operator()(int);
   const ScalarType& operator()(int) const;
   friend void swap(Container<Scalar,0>& A_c1, Container<Scalar,0>& A_c2)
   {
       std::swap(A_c1.data,A_c2.data);
       std::swap(A_c1.capacity,A_c2.capacity);
   }

   static void* operator new[](size_t sizeOfData, int);
   static void* operator new[](size_t sizeOfData);
   static void setDefaultDimension(int newDefaultDim)
   {
      defaultDimension = newDefaultDim;
   }
   const int dimension() const {return capacity;}


protected:
   ScalarType *data;
   int capacity;
   static int defaultDimension;
};

// --------------- iterator selection --------------------- //

template<typename Scalar>
inline typename Container<Scalar,0>::iterator Container<Scalar,0>::begin()
{
   return iterator(data);
}

template<typename Scalar>
inline typename Container<Scalar,0>::iterator Container<Scalar,0>::end()
{
   return iterator(data+capacity);
}

template<typename Scalar>
inline typename Container<Scalar,0>::const_iterator Container<Scalar,0>::begin() const
{
   return const_iterator(data);
}

template<typename Scalar>
inline typename Container<Scalar,0>::const_iterator Container<Scalar,0>::end() const
{
   return const_iterator(data+capacity);
}

// ------------------ indexing --------------------------- //

template<typename Scalar>
inline Scalar& Container<Scalar,0>::operator[] (int i)
{
   return data[i];
}

template<typename Scalar>
inline const Scalar& Container<Scalar,0>::operator[] (int i) const
{
   return data[i];
}

template<typename Scalar>
inline Scalar& Container<Scalar,0>::operator() (int i)
{
   return data[i-1];
}

template<typename Scalar>
inline const Scalar& Container<Scalar,0>::operator() (int i) const
{
   return data[i-1];
}

// ----------------- memory allocation ------------------------

template <typename Scalar>
inline void* Container<Scalar,0>::operator new[](size_t sizeOfData)
{
   return ::operator new[](sizeOfData);
}

template <typename Scalar>
inline void* Container<Scalar,0>::operator new[](size_t sizeOfData,int newDim)
{
   defaultDimension = newDim;
   return ::operator new[](sizeOfData);
}

// --------------- member definitions ----------------------------- //

template<typename Scalar>
inline Container<Scalar,0>::~Container()
{
   delete [] data;
}

template<typename Scalar>
inline Container<Scalar,0>::Container() : data(0), capacity(defaultDimension)
{
   if(capacity)
   {
      data = new ScalarType[capacity];
      iterator b=begin(), e=end();
      while(b!=e)
      {
         *b=ScalarType(0);
         ++b;
      }
   }
}

template<typename Scalar>
inline Container<Scalar,0>::Container(int a_capacity) : capacity(a_capacity)
{
   data = new ScalarType[capacity];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b=ScalarType(0);
      ++b;
   }
}

template<typename Scalar>
inline Container<Scalar,0>::Container(int a_capacity,bool) : capacity(a_capacity)
{
   data = new ScalarType[capacity];
}

template<typename Scalar>
inline Container<Scalar,0>::Container(const Container& a_container) : capacity(a_container.capacity)
{
   data = new ScalarType[capacity];
   iterator b=begin(), e=end();
   const_iterator i = a_container.begin();
   while(b!=e)
   {
      *b = *i;
      ++b;
      ++i;
   }
}

template<typename Scalar>
Container<Scalar,0>& Container<Scalar,0>::operator=(const Container& a_c)
{
   if(capacity!=a_c.capacity)
   {
      delete [] data;
      capacity =  a_c.capacity;
      data = new ScalarType[capacity];
   }

   iterator b=begin(), e=end();
   const_iterator i = a_c.begin();
   while(b!=e)
   {
      *b = *i;
      ++b;
      ++i;
   }

   return *this;
}

template<typename Scalar>
void Container<Scalar,0>::resize(int A_newCapacity)
{
   if(capacity!=A_newCapacity)
   {
      if(data) delete[] data;
      capacity = A_newCapacity;
      data = new ScalarType[capacity];
   }
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = Scalar(0);
      ++b;
   }
}


template<typename Scalar>
int Container<Scalar,0>::defaultDimension = 0;

}} // namespace capd::vectalg

#endif

/// @}
