/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file matrixintf.h
///
/// @author The CAPD group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59

#ifndef _matrixTemplate
#define _matrixTemplate

#include <iostream>
#include "vectalg/container.h"
#include "vectalg/matrixContainer.h"
#include "vectalg/matrixIterator.h"
#include "vectalg/matrixSlice.h"
#include "vectalg/rowVector.h"
#include "vectalg/columnVector.h"

template<typename Scalar,int dim>
class Vector;

template<typename Scalar,int rows, int cols>
class Matrix;

template<typename Scalar,int rows, int cols>
Matrix<Scalar,rows,cols> abs(const Matrix<Scalar,rows,cols>&);

template<typename Scalar,int rows, int cols>
Matrix<Scalar,rows,cols> operator-(const Matrix<Scalar,rows,cols>&);

template<typename Scalar,int rows, int cols>
Matrix<Scalar,rows,cols> operator+(const Matrix<Scalar,rows,cols>&, const Matrix<Scalar,rows,cols>&);

template<typename Scalar,int rows, int cols>
Matrix<Scalar,rows,cols> operator-(const Matrix<Scalar,rows,cols>&, const Matrix<Scalar,rows,cols>&);

template<typename Scalar, int rows,int cols>
Matrix<Scalar,rows,cols> operator*(const Matrix<Scalar,rows,cols>&, const Scalar&);

template<typename Scalar, int rows,int cols1, int cols2>
Matrix<Scalar,rows,cols2> operator*(const Matrix<Scalar,rows,cols1>&, const Matrix<Scalar,cols1,cols2>&);

template<typename Scalar, int rows,int cols>
Matrix<Scalar,rows,cols> operator*(const Scalar&, const Matrix<Scalar,rows,cols>&);

template<typename Scalar,int rows,int cols>
Vector<Scalar,rows> operator*(const Matrix<Scalar,rows,cols>&, const Vector<Scalar,cols>&);

template<typename Scalar, int rows,int cols>
Matrix<Scalar,rows,cols> operator/(const Matrix<Scalar,rows,cols>&, const Scalar&);

template<typename Scalar, int rows,int cols>
Matrix<Scalar,rows,cols> operator+(const Matrix<Scalar,rows,cols>&, const Scalar&);

template<typename Scalar, int rows,int cols>
Matrix<Scalar,rows,cols> operator-(const Matrix<Scalar,rows,cols>&, const Scalar&);

template<typename Scalar, int rows,int cols>
bool operator<(const Matrix<Scalar,rows,cols>&, const Matrix<Scalar,rows,cols>&);

template<typename Scalar, int rows,int cols>
bool operator>(const Matrix<Scalar,rows,cols>&, const Matrix<Scalar,rows,cols>&);

template<typename Scalar, int rows,int cols>
bool operator<=(const Matrix<Scalar,rows,cols>&, const Matrix<Scalar,rows,cols>&);

template<typename Scalar, int rows,int cols>
bool operator>=(const Matrix<Scalar,rows,cols>&, const Matrix<Scalar,rows,cols>&);

template<typename Scalar, int rows,int cols>
Matrix<Scalar,cols,rows> Transpose(const Matrix<Scalar,rows,cols>&);

template<typename Scalar, int rows,int cols>
std::ostream &operator<<(std::ostream&, const Matrix<Scalar,rows,cols>&);

template<typename Scalar, int rows,int cols>
std::istream &operator>>(std::istream&, Matrix<Scalar,rows,cols>&);

/*  *** MM ***
template<typename Scalar>
void swap(Matrix<Scalar,0,0>& A_m1, Matrix<Scalar,0,0>& A_m2);

*/
// ------------------------------------------------------------------------------//

template<typename IMatrixType>
void splitMatrix(IMatrixType& v, IMatrixType& ra);

template<typename IMatrixType>
IMatrixType midMatrix(const IMatrixType& v);

// ########################################################################### //

template<typename Scalar, int rows, int cols>
class Matrix : public capd::vectalg::MatrixContainer<Scalar,rows,cols>
{
public:
   typedef Scalar ScalarType;
   typedef capd::vectalg::MatrixContainer<Scalar,rows,cols> ContainerType;
   typedef typename ContainerType::iterator ContainerIterator;
   typedef typename ContainerType::const_iterator const_ContainerIterator;
   typedef Vector<Scalar,cols> RowVectorType;
   typedef Vector<Scalar,rows> ColumnVectorType;
   typedef RowVector<Scalar,cols> RefRowVectorType;
   typedef ColumnVector<Scalar,rows> RefColumnVectorType;
   typedef Matrix<Scalar,rows,cols> MatrixType;

   //constructors
   Matrix();
   Matrix(int A_rows,int A_cols);               //assigns 0 to every coordinate - maybe shoudn't?
   Matrix(const ScalarType data[]);
   Matrix(int A_rows,int A_cols, const ScalarType data[]);
   Matrix(const Matrix& m);                     // copying constructor
   Matrix(const MatrixSlice<MatrixType>& m);    // copying from matrix slice
   Matrix(Matrix<double,rows,cols>&);
   Matrix(int A_rows,int A_cols,bool) : ContainerType(A_rows,A_cols,true){}
   void clear();                                //assigns zero to each coord

   //assignments - matrices
   Matrix& operator= (const Matrix& a);         //assign a matrix
   Matrix& operator+=(const Matrix& a);         //increase by a matrix
   Matrix& operator-=(const Matrix& a);         //decrease by a matrix

   //assignments - Scalars
   Matrix& operator= (const Scalar& s);         //assign a Scalar
   Matrix& operator+=(const Scalar& s);         //increase by a Scalar
   Matrix& operator-=(const Scalar& s);         //decrease by a Scalar
   Matrix& operator*=(const Scalar& s);         //rescale by multiplying
   Matrix& operator/=(const Scalar& s);         //rescale by dividing

   // iterator selections
   MatrixIterator< MatrixType > begin();
   MatrixIterator< MatrixType > end();
   MatrixIterator< MatrixType > beginOfRow(int i);
   MatrixIterator< MatrixType > beginOfColumn(int j);
   MatrixIterator< MatrixType > endOfRow(int i);
   MatrixIterator< MatrixType > endOfColumn(int j);

   const_MatrixIterator< MatrixType > begin() const;
   const_MatrixIterator< MatrixType > end() const;
   const_MatrixIterator< MatrixType > beginOfRow(int i) const;
   const_MatrixIterator< MatrixType > beginOfColumn(int j) const;
   const_MatrixIterator< MatrixType > endOfRow(int i) const;
   const_MatrixIterator< MatrixType > endOfColumn(int j) const;

   inline ContainerIterator cbegin(){ return ContainerType::begin();}
   inline ContainerIterator cend()  { return ContainerType::end();  }
   inline const_ContainerIterator cbegin() const { return ContainerType::begin();}
   inline const_ContainerIterator cend() const   { return ContainerType::end();  }

   //indexing
   RowVector<Scalar,cols> operator[](int i) const; // i-th row as a vector
   RowVector<Scalar,cols> operator()(int i) const; // (i-1)-th row as a vector
   Scalar& operator()(int i,int j);                 //returns reference to the [i-1][j-1] entry
   const Scalar& operator()(int i,int j) const;     //returns const reference to the [i-1][j-1] entry
   Scalar* at(int i,int j);                         //returns pointer to the [i-1][j-1] entry
   const Scalar* at(int i,int j) const;             //returns pointer to the [i-1][j-1] entry

   //operations on matrices
   static Matrix Identity(int dimension);              // returns the identity matrix
   void setToIdentity();                               // if square matrix, changes it to the identity matrix
   RowVector<Scalar,cols> row(int i) const;       // i-th row as a vector
   ColumnVector<Scalar,rows> column(int j) const; // returns j-th column
   void Transpose();

   const int rowStride()const {
      return numberOfColumns();
   } // difference of pointers when moving to the next row in the same column
   const int columnStride()const{
      return 1;
   }  // difference of pointers when moving to the next column in the same row

   using ContainerType::numberOfRows;
   using ContainerType::numberOfColumns;
//   using ContainerType::operator new[];

   friend std::istream &operator>> <> (std::istream &inp, MatrixType &a);

protected:
   using ContainerType::resize;
   using ContainerType::data;
}; //the end of class Matrix

// ###################### inline definitions ############################# //

template<typename Scalar, int rows, int cols>
inline Matrix<Scalar,rows,cols>::Matrix() : ContainerType()
{}

template<typename Scalar, int rows, int cols>
inline Matrix<Scalar,rows,cols>::Matrix(int A_rows,int A_cols) : ContainerType(A_rows,A_cols)
{}

template<typename Scalar, int rows, int cols>
inline Matrix<Scalar,rows,cols>::Matrix(const Matrix& a_m) : ContainerType(a_m)
{}


template<typename Scalar, int rows, int cols>
inline Matrix<Scalar,rows,cols>& Matrix<Scalar,rows,cols>::operator=(const Matrix& a)
{
   ContainerType::operator=(a);
   return *this;
}


template<typename Scalar, int rows, int cols>
inline RowVector<Scalar,cols> Matrix<Scalar,rows,cols>::operator[](int i) const
{
   return RowVector<Scalar,cols>(data+i*numberOfColumns(),numberOfColumns());
}


template<typename Scalar, int rows, int cols>
inline RowVector<Scalar,cols> Matrix<Scalar,rows,cols>::operator()(int i) const
{
   return RowVector<Scalar,cols>(data + --i*numberOfColumns(),numberOfColumns());
}

template<typename Scalar, int rows, int cols>
inline RowVector<Scalar,cols> Matrix<Scalar,rows,cols>::row(int i) const
{
   return RowVector<Scalar,cols>(data+i*numberOfColumns(),numberOfColumns());
}

template<typename Scalar, int rows, int cols>
inline ColumnVector<Scalar,rows> Matrix<Scalar,rows,cols>::column(int i) const
{
   return ColumnVector<Scalar,rows>(data+i,numberOfColumns(),numberOfRows());
}

template<typename Scalar, int rows, int cols>
inline Scalar& Matrix<Scalar,rows,cols>::operator()(int i,int j)
{
   return data[ --i*numberOfColumns() + --j];
}

template<typename Scalar, int rows, int cols>
inline const Scalar& Matrix<Scalar,rows,cols>::operator()(int i,int j) const
{
   return data[ --i*numberOfColumns() + --j];
}

template<typename Scalar, int rows, int cols>
inline Scalar* Matrix<Scalar,rows,cols>::at(int i,int j)
{
   return data + --i*numberOfColumns() + --j;
}

template<typename Scalar, int rows, int cols>
inline const Scalar* Matrix<Scalar,rows,cols>::at(int i,int j) const
{
   return data + --i*numberOfColumns() + --j;
}

// ------------------------ iterators ----------------------------- //

template<typename Scalar, int rows, int cols>
inline MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::begin()
{
  return MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(1,1));
}

template<typename Scalar, int rows, int cols>
inline MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::end()
{
  return MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(numberOfRows(),numberOfColumns()+1));
}

template<typename Scalar, int rows, int cols>
inline MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::beginOfRow(int i)
{
  return MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(i,1));
}

template<typename Scalar, int rows, int cols>
inline MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::beginOfColumn(int j)
{
  return MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(1,j));
}

template<typename Scalar, int rows, int cols>
inline MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::endOfRow(int i)
{
  return MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(i,numberOfColumns()+1));
}

template<typename Scalar, int rows, int cols>
inline MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::endOfColumn(int j)
{
  return MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(numberOfRows(),j+1));
}


template<typename Scalar, int rows, int cols>
inline const_MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::begin() const
{
  return const_MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(1,1));
}

template<typename Scalar, int rows, int cols>
inline const_MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::end() const
{
  return const_MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(numberOfRows(),numberOfColumns()+1));
}

template<typename Scalar, int rows, int cols>
inline const_MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::beginOfRow(int i) const
{
  return const_MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(i,1));
}

template<typename Scalar, int rows, int cols>
inline const_MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::beginOfColumn(int j) const
{
  return const_MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(1,j));
}

template<typename Scalar, int rows, int cols>
inline const_MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::endOfRow(int i) const
{
  return const_MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(i,numberOfColumns()+1));
}

template<typename Scalar, int rows, int cols>
inline const_MatrixIterator< Matrix<Scalar,rows,cols> > Matrix<Scalar,rows,cols>::endOfColumn(int j) const
{
  return const_MatrixIterator< Matrix<Scalar,rows,cols> >(*this,at(numberOfRows(),j+1));
}

#endif

/// @}
